
from elevenlabs import generate, save, set_api_key
import os
from dotenv import load_dotenv

load_dotenv()

API_KEY = os.getenv('ELEVENLABS_API_KEY')

# TODO: Implement a method to get an existing voice attending to a 'type' (terror, inspirational, etc.)

def narrate(text, voice, output_filename):
    """
    Receives a 'text' and generates a single audio file with that 'text' narrated with
    the provided 'voice', stored locally as 'output_filename'.

    This method will split 'text' if too much longer to be able to narrate without issues
    due to external platform working process. But will lastly generate a single audio file.
    """
    texts = [text]
    # TODO: Set this limit according to voice type
    if len(text) > 999999:
        texts = []
        # TODO: Handle splitting text into subgroups to narrate and then join
        print('No subgrouping text yet')
        texts = [text]

    if len(texts) == 1:
        # Only one single file needed
        download_elevenlabs_audio(texts[0], voice, output_filename)
    else:
        for text in texts:
            # TODO: Generate single file
            print('Not implemented yet')

        # TODO: Join all generated files in only one (maybe we need some silence in between?)
            
    return output_filename

def download_elevenlabs_audio(text = 'Esto es API', voice = 'Freya', output_file = 'generated_elevenlabs.wav'):
    """
    Generates a narration in elevenlabs and downloads it as output_file audio file.
    """
    set_api_key(API_KEY)
    # TODO: Check if voice is valid
    # TODO: Check which model fits that voice.
    model = 'eleven_multilingual_v2'

    if not output_file.endswith('.wav'):
        output_file = output_file + '.wav'

    # TODO: Try to be able to call it with stability parameter
    audio = generate(
        text = text,
        voice = voice,
        model = model
    )

    save(audio, output_file)
    


from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By

import time
import os
import pyautogui
import glob
import shutil

PROJECT_ABSOLUTE_PATH = os.getenv('PROJECT_ABSOLUTE_PATH')
TOUSE_ABSOLUTE_PATH = os.getenv('TOUSE_ABSOLUTE_PATH') 

def go_to_and_wait_loaded(driver, url):
    """
    Navigates to the provided 'url', using the provided 'driver', and waits until
    the page is loaded (by readyState method). If it is loaded, it returns True.
    If not, it returns False.

    This method will wait for 10 seconds maximum.
    """
    try:
        driver.get(url)

        page_state = 'loading'
        cont = 0
        while page_state != 'complete' and cont < 20:
            time.sleep(0.5)
            page_state = driver.execute_script('return document.readyState;')
            cont += 1

        if page_state == 'loading':
            return False
        
        return True
    except:
        driver.close()

def start_chrome(gui = False, ad_blocker = True, disable_popups_and_cookies = True, additional_options = []):
    """
    Starts the chrome driver with the local configuration and returns it. If 'gui' is True,
    it will show the Chrome working.

    Additional options must be simple strings. Cookies must be objects with .name and .value.
    """
    try:
        CHROME_EXTENSIONS_ABSOLUTEPATH = 'C:/Users/dania/AppData/Local/Google/Chrome/User Data/Profile 2/Extensions/'
        AD_BLOCK_ABSOLUTEPATH = CHROME_EXTENSIONS_ABSOLUTEPATH + 'cjpalhdlnbpafiamejdnhcphjbkeiagm/1.56.0_0'

        # TODO: What is 'import undetected_chromedriver.v2 as uc'? Try it

        options = Options()
        
        option_arguments = ['--start-maximized']

        if len(additional_options) > 0:
            for additional_option in additional_options:
                option_arguments.append(additional_option)

        if not gui:
            option_arguments.append('--headless=new')

        if ad_blocker:
            # This loads the ad block 'uBlock' extension that is installed in my pc
            option_arguments.append('load-extension=' + AD_BLOCK_ABSOLUTEPATH)
        
        # Load user profile
        option_arguments.append('user-data-dir=C:/Users/dania/AppData/Local/Google/Chrome/User Data/Profile 2')

        for argument in option_arguments:
            options.add_argument(argument)

        if disable_popups_and_cookies:
            # TODO: Separate this into specific options, not all together. One is for cookies,
            # another one is for popups... Separate them, please
            # This disables popups, cookies and that stuff
            options.add_experimental_option("prefs", {"excludeSwitches": ['enable-automation', 'load-extension', 'disable-popup-blocking'], 'profile.default_content_setting_values.automatic_downloads': 1})

        driver = webdriver.Chrome(options = options)

        return driver
    except:
        driver.close()

def download_fake_call_image(name, output_filename):
    # TODO: Move this to the faker
    URL = 'https://prankshit.com/fake-iphone-call.php'

    try:
        driver = start_chrome()
        go_to_and_wait_loaded(driver, URL)

        inputs = driver.find_elements(By.TAG_NAME, 'input')
        name_textarea = driver.find_element(By.TAG_NAME, 'textarea')

        #operator_input = inputs[4]
        #hour_input = inputs[5]

        name_textarea.clear()
        name_textarea.send_keys(name)

        image = driver.find_element(By.XPATH, '//div[contains(@class, "modal-content tiktok-body")]')
        image.screenshot(output_filename)
    finally:
        driver.close()

# Other fake generators (https://fakeinfo.net/fake-twitter-chat-generator) ad (https://prankshit.com/fake-whatsapp-chat-generator.php)
def download_discord_message_image(text, output_filename):
    URL = 'https://message.style/app/editor'

    try:
        driver = start_chrome()
        go_to_and_wait_loaded(driver, URL)
        
        time.sleep(3)

        clear_embed_button = driver.find_element(By.XPATH, "//button[contains(text(), 'Clear Embeds')]")
        clear_embed_button.click()

        time.sleep(3)

        input_elements = driver.find_elements(By.TAG_NAME, 'input')
        username_input = input_elements[3]
        avatar_url_input = input_elements[4]

        username_input.clear()
        username_input.send_keys('botsito')

        avatar_url_input.clear()
        avatar_url_input.send_keys('https://cdn.pixabay.com/photo/2016/11/18/23/38/child-1837375_640.png')

        textarea_input = driver.find_element(By.TAG_NAME, 'textarea')
        textarea_input.clear()
        textarea_input.send_keys(text)

        time.sleep(3)

        # get element div class='discord-message'
        discord_message = driver.find_element(By.XPATH, "//div[contains(@class, 'discord-message')]")
        discord_message.screenshot(output_filename)
    finally:
        driver.close()


def navigate_to_and_screenshot(url, duration):
    """
    Browses the website 'url' and takes screenshots  while scrolling down. This method
    takes 30 screenshots per second.

    Returns the array of screenshot filenames stored locally.
    """
    try:
        FPS = 30

        driver = start_chrome()
        go_to_and_wait_loaded(driver, url)
        
        page_height = 0
        page_height = driver.execute_script("return document.body.scrollHeight")
        print(page_height)
        # We force it to 1000 if bigger
        if page_height > 1000:
            page_height = 1000

        screenshot_filenames = []
        number_of_screenshots = int(duration * FPS)

        # How much should we scroll?
        new_height = 0
        for i in range(number_of_screenshots):
            screenshot_filename = 'wip/screenshot_' + str(i) + '.png'
            driver.save_screenshot(screenshot_filename)
            screenshot_filenames.append(screenshot_filename)
            height = driver.execute_script("return window.pageYOffset")
            new_height += page_height / number_of_screenshots
            driver.execute_script('window.scrollTo(' + str(height) + ', ' + str(new_height) + ')')

        driver.execute_script("window.close();")
    finally:
        driver.close()

    # TODO: These screenshots are not 1920x1080, so crop them from center to sides
    return screenshot_filenames
    

def test_download_piano_music():
    # TODO: End this, to make it download music generated by this AI
    try:
        options = Options()
        #option_arguments = ['--start-maximized', '--headless=new']
        option_arguments = ['--start-maximized']
        for argument in option_arguments:
            options.add_argument(argument)

        driver = webdriver.Chrome(options = options)
        driver.get('https://huggingface.co/spaces/mrfakename/rwkv-music')
        wait = WebDriverWait(driver, 30)
        download_button_element = wait.until(EC.visibility_of_element_located((By.XPATH, '//*[@data-testid=\"checkbox\"]')))

        time.sleep(5)
        actions = ActionChains(driver)
        for i in range(11):
            actions.send_keys(Keys.TAB)
        actions.perform()
        actions.send_keys(Keys.SPACE)
        actions.perform()

        input_number_element = driver.find_elements_by_xpath('//*[@data-testid="number-input"]')[0]
        input_number_element.send_keys(14286)

        actions = ActionChains(driver)
        actions.send_keys(Keys.TAB)
        actions.perform()

        time.sleep(1)

        driver.execute_script('window.scrollTo(0, 1000)')

        time.sleep(4)

        download_button_element = wait.until(EC.visibility_of_element_located((By.XPATH, '//*[@title=\"Download\"]')))
        #download_button_element = driver.find_elements_by_xpath('//*[@title="Download"]')[0]
        download_button_element.click()

    finally:
        driver.close()


def goto_and_scroll(url):
    # Intersting options: https://github.com/ultrafunkamsterdam/undetected-chromedriver/issues/1726
    # Also this: https://stackoverflow.com/questions/19211006/how-to-enable-cookies-in-chromedriver-with-webdriver
    # What about this: options.AddUserProfilePreference("profile.cookie_controls_mode", 0);
    try:
        options = Options()
        option_arguments = ['--start-maximized', '--disabled-infobars', '--disable-blink-features=AutomationControlled', '--disable-animations', '--no-sandbox', '--disable-setuid-sandbox']
        # options.AddUserProfilePreference("profile.cookie_controls_mode", 0);
        for argument in option_arguments:
            options.add_argument(argument)

        """
        experimental_option_arguments = [{
            'option': 'excludeSwitches',
            'value': 'enable-automation'
        }]
        if len(experimental_option_arguments) > 0:
            options.add_experimental_option(experimental_option_arguments)
        """
        # Remove this line below for debug
        #options.add_argument("--headless=new") # for Chrome >= 109
        print('@@@@@@@@@@@@@')
        driver = webdriver.Chrome(options = options)
        print('@@@@@@@@@@@@@')
        driver.get(url)
        print('here i am')

        # Accept cookies

        #driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
        time.sleep(4)

        driver.execute_script('window.scrollTo(0, 200)')
        time.sleep(4)
    except Exception as e:
        print(e)
    finally:
        driver.close()

def get_redicted_url(url, expected_url = None, wait_time = 5):
    """
    Navigates to the provided url and waits for a redirection. This method will wait
    until the 'expected_url' is contained in the new url (if 'expected_url' parameter
    is provided), or waits 'wait_time' seonds to return the current_url after that.
    """
    redirected_url = ''

    try:
        options = Options()
        options.add_argument("--start-maximized")
        # Remove this line below for debug
        options.add_argument("--headless=new") # for Chrome >= 109
        driver = webdriver.Chrome(options = options)
        driver.get(url)

        wait = WebDriverWait(driver, 10)

        if not expected_url:
            time.sleep(wait_time)
        else:
            wait.until(EC.url_contains(expected_url))

        redirected_url = driver.current_url
    finally:
        driver.close()

    return redirected_url


def get_youtube_summary(video_id):
    """
    Searchs into 'summarize.tech' web to obtain the summary of the video with the 
    provided 'video_id'. This method returns the summary in English, as it is 
    provided by that website.
    """
    url = 'https://www.summarize.tech/www.youtube.com/watch?v=' + video_id

    try:
        options = Options()
        options.add_argument("--start-maximized")
        # Remove this line below for debug
        options.add_argument("--headless=new") # for Chrome >= 109
        driver = webdriver.Chrome(options = options)
        driver.get(url)

        summary = driver.find_element_by_tag_name('section').find_element_by_tag_name('p').get_attribute('innerText')
    finally:
        driver.close()

    return summary

def google_translate(text, input_language = 'en', output_language = 'es'):
    url = 'https://translate.google.com/?hl=es'
    """
    https://translate.google.com/?hl=es&sl=en&tl=es&text=Aporta%20una%20unidad%20de%20traducci%C3%B3n%20(segmento%20y%20traducci%C3%B3n)%20en%20alg%C3%BAn%20par%20de%20idiomas%20a%20MyMemory.%0ASin%20especificar%20ning%C3%BAn%20par%C3%A1metro%20clave%2C%20la%20contribuci%C3%B3n%20est%C3%A1%20disponible%20para%20todos%20(%C2%A1Gracias!).&op=translate

    https://translate.google.com/?hl=es&tab=TT&sl=en&tl=es&op=translate

    https://translate.google.com/?hl=es&tab=TT&sl=en&tl=es&text=La%20%C3%BAnica%20forma%20de%20saberlo%20es%20lo%20que%20t%C3%BA%20digas&op=translate
    """

    url = 'https://translate.google.com/?hl=' + output_language + '&tab=TT&sl=' + input_language + '&tl=' + output_language + '&text=' + text + '&op=translate'

    translation = ''

    try:
        options = Options()
        options.add_argument("--start-maximized")
        # Comment this line below for debug (enables GUI)
        options.add_argument("--headless=new") # for Chrome >= 109
        driver = webdriver.Chrome(options = options)
        driver.get(url)

        tries = 0
        while True:
            if tries < 20:
                try:
                    # We try until it doesn't fail (so we have the text)
                    # TODO: This 'jscontroller' changes from time to time, pay atention
                    translation = driver.find_elements('xpath', '//*[@jscontroller="JLEx7e"]')[0].get_attribute('innerText')
                    tries = 20
                except Exception as e:
                    # TODO: Uncomment this to see if code error or scrapper error
                    #print(e)
                    tries += 1
                    time.sleep(0.250)
            else:
                break
    finally:
        driver.close()

    return translation


from gtts import gTTS
from moviepy.editor import AudioFileClip, CompositeAudioClip
from pydub import AudioSegment, silence
from df.enhance import enhance, init_df, load_audio, save_audio

import pyttsx3
import os

def generate_cortana_narration(text, output_filename):
    """
    Creates an audio narration of the provided 'text' and stores it as 'output_filename'.
    """
    engine = pyttsx3.init()
    engine.save_to_file(text, output_filename)
    engine.runAndWait()

def generate_google_narration(text, output_filename, language = 'es'):
    """
    Creates an audio narration of the provided 'text' with the Google voice and stores it
    as 'output_filename'. This will use the provided 'language' language for the narration.
    """
    # TODO: Check valid language tag in this table (https://en.wikipedia.org/wiki/IETF_language_tag)
    # TODO: Use this library for languages (https://pypi.org/project/langcodes/)
    tts = gTTS(text, lang = language)
    tts.save(output_filename)

def crop(audio_filename, duration, output_filename):
    """
    Crops the 'audio_filename' provided to the requested 'duration'.

    This method returns the new audio 'output_filename' if valid, or
    False if it was not possible to crop it.
    """
    audio_clip = AudioFileClip(audio_filename)

    if audio_clip.duration < duration:
        print('audio is shorter than requested duration')
        return False
    
    audio_clip.set_duration(duration).write_audiofile(output_filename)

    return output_filename

def mp3_to_wav(mp3_filename, output_wav_filename):
    """
    Receives an .mp3 file 'mp3_filename' and turns it into a .wav
    file stored in 'output_wav_filename'.
    """
    # TODO: Generalize this method to swap between formats
    sound = AudioSegment.from_mp3(mp3_filename)
    sound.export(output_wav_filename, format = "wav")

def remove_noise(audio_filename, audio_output_filename):
    """
    Removes the noise from the provided 'audio_filename' and creates a new
    file 'audio_output_filename' without noise.

    TODO: This fails when .mp3 is used, so we need to transform into wav.

    TODO: Output file must be also wav
    """
    # Based on this (https://medium.com/@devesh_kumar/how-to-remove-noise-from-audio-in-less-than-10-seconds-8a1b31a5143a)
    # https://github.com/Rikorose/DeepFilterNet
    # Guide to installation is right here, in the link below:
    # https://github.com/Rikorose/DeepFilterNet?tab=readme-ov-file#deepfilternet-python-pypi
    # TODO: This is failing now saying 'File contains data in an unknon format'...
    # I don't know if maybe some library, sh*t...
    # Load default model
    TMP_WAV_FILENAME = 'tmp_wav.wav'
    if audio_filename.endswith('.mp3'):
        # TODO: Maybe it is .wav but not that format...
        mp3_to_wav(audio_filename, TMP_WAV_FILENAME)
        audio_filename = TMP_WAV_FILENAME

    model, df_state, _ = init_df()
    audio, _ = load_audio(audio_filename, sr = df_state.sr())
    # Remove the noise
    enhanced = enhance(model, df_state, audio)

    save_audio(audio_output_filename, enhanced, df_state.sr())

    try:
        os.remove(TMP_WAV_FILENAME)
    except:
        pass

def get_audio_length_in_seconds(input_file):
    """
    Returns the provided audio file 'input_file' duration in seconds.

    This method uses AudioFileClip.duration.
    
    It must be an audio file.
    """
    audiofileclip = AudioFileClip(input_file)
    
    return audiofileclip.duration

def add_audio_to_audio(audio_filename, new_audio_filename, start_time = 1.0):
    """
    We receive an audio file called 'audio_filename' and we append the provided
    'new_audio_filename' to that existing audio to sound over it. That second 
    sound will start in the 'start_time' second of the first 'audio_filename'
    """
    first_audioclip = AudioFileClip(audio_filename)
    second_audioclip = AudioFileClip(new_audio_filename)

    combined = CompositeAudioClip([first_audioclip, second_audioclip.set_start(start_time)])
    combined.fps = 44100
    
    combined.write_audiofile('audio_combined.mp3')

def detect_silences(audio_filename, silence_min_length = 250):
    """
    This method detects the silences existing in the provided 'audio_filename' that
    are of a minimum of 'silence_min_length' long. It will return an array containing
    tuples (X.XX, Y.YY) of start and end silence moments (in seconds).
    """
    audio = __get_audio_segment(audio_filename)

    dBFS = audio.dBFS
    silences = silence.detect_silence(audio, min_silence_len = silence_min_length, silence_thresh = dBFS - 16)

    # [(1.531, 1.946), (..., ...), ...] in seconds
    return [((start / 1000), (stop / 1000)) for start, stop in silences]

def __get_audio_segment(audio_filename):
    if not '.' in audio_filename:
        return None
    
    extension = audio_filename.split('.')[1]
    myaudio = AudioSegment.from_file(audio_filename)
    """
    if extension == 'mp3':
        myaudio = AudioSegment.from_mp3(audio_filename)
    elif extension == 'wav':
        myaudio = AudioSegment.from_wav(audio_filename)
    else:
        # TODO: Do this work? I think no, because it is not configured
        myaudio = AudioSegment.from_file(audio_filename)
    """

    return myaudio


import requests

def download_file(url, output_filename):
    """
    Receives a downloadable url as 'url' and downloads that file in
    our system as 'output_filename'.
    """
    r = requests.get(url)

    with open(output_filename, 'wb') as outfile:
        outfile.write(r.content)

    return True

def download_audio(url, output_filename):
    """
    Receives a downloadable url as 'url' and downloads that audio in
    our system as 'output_filename'.
    """
    # TODO: Check formats and that it is a valid audio (and use Enums pls)
    if not output_filename.endswith('.mp3') and not output_filename.endswith('.wav') and not output_filename.endswith('.m4a'):
        output_filename = output_filename + '.mp3'

    return download_file(url, output_filename)

def download_image(url, output_filename):
    """
    Receives a downloadable url as 'url' and downloads that image in
    our system as 'output_filename'.
    """
    # TODO: Check formats and that it is a valid image (and use Enums pls)
    if not output_filename.endswith('.jpg') and not output_filename.endswith('.png') and not output_filename.endswith('.jpeg') and not output_filename.endswith('.webp') and not output_filename.endswith('.bmp'):
        output_filename = output_filename + '.jpg'

    return download_file(url, output_filename)

def download_video(url, output_filename):
    """
    Receives a downloadable url as 'url' and downloads that video in
    our system as 'output_filename'.
    """
    # TODO: Check formats and that it is a valid audio (and use Enums pls)
    if not output_filename.endswith('.mp4') and not output_filename.endswith('.mov'):
        output_filename = output_filename + '.mp4'

    return download_file(url, output_filename)


"""
    TODO:

    This must be deleted, but I keep this code because it is working.
    This is not dynamic nor personalized, so it won't fit this 
    library because it is not reusable.

    Please, delete this when you find a way of having reusable
    code, thank you.
"""


import mysql.connector
import os
from dotenv import load_dotenv

load_dotenv()

# MYSQL
"""
mydb = mysql.connector.connect(
  host = os.getenv('DATABASE_HOST'),
  user = os.getenv('DATABASE_USER'),
  password = os.getenv('DATABASE_PASSWORD'),
  database = os.getenv('DATABASE_NAME')
)

mycursor = mydb.cursor(buffered = True)
"""
mydb = None
mycursor = None

def __database_video_to_dict(database_row):
    """
    Receives a database row format (columns index) and returns it
    as a dictionary to make manipulation easier
    """
    return {
        'id': database_row[0],
        'video_id': database_row[1],
        'name': database_row[2],
        'description': database_row[3],
        'is_uploaded': database_row[4]
    }

def get_video_from_database(table, video_id):
    """
    Looks for the video with 'video_id' provided in database and
    returns None if it doesn't exist or a dictionary if it does
    """
    sql = "SELECT * FROM " + table + " WHERE video_id = '" + video_id + "' LIMIT 1"
    mycursor.execute(sql)

    if mycursor.rowcount == 0:
        return None
    
    return __database_video_to_dict(mycursor.fetchall()[0])

def insert_video(table, video_id, name, description, is_uploaded):
    """
    Inserts a new video row in our database
    """
    # TODO: Sorry, but this changes depending on the object
    sql = "INSERT INTO " + table + " (video_id, name, description, is_uploaded) VALUES (%s, %s, %s, %s)"
    val = (video_id, name, description, is_uploaded)
    mycursor.execute(sql, val)
    mydb.commit()

def db_insert(table, fields, values):
    # TODO: This 'VALUES' is not dynamic
    sql = "INSERT INTO " + table + " (" + ', '.join(fields) + ") VALUES (%s, %s, %s)"
    mycursor.execute(sql, tuple(values))
    mydb.commit()

def db_select(table, field, value):
    sql = "SELECT * FROM " + table + " WHERE " + field + " = '" + value + "' LIMIT 1"
    mycursor.execute(sql)

    if mycursor.rowcount == 0:
        return None
    
    result = mycursor.fetchall()[0]

    # TODO: Sorry, but this changes depending on the object
    return {
        'id': result[0],
        'video_id': result[1],
        'username': result[2],
        'status': result[3],
    }




import requests

def url_is_ok(url):
    """
    Checks if the provided url is valid. It returns True if yes or
    False if not.
    """
    try:
        response = requests.head(url)

        if response.status_code == 200:
            return True
        else:
            return False
    except requests.ConnectionError as e:
        return e
    
def url_is_image(url):
    """
    Checks if the provided url is valid and is a valid image. It returns
    True if yes or False if not.
    """
    # TODO: Add more formats
    image_formats = ("image/png", "image/jpeg", "image/jpg")
    r = requests.head(url)
    if r.headers["content-type"] in image_formats:
        return True
    return False


from moviepy.editor import VideoFileClip
from math import floor
from dotenv import load_dotenv

load_dotenv()

WIP_FOLDER = os.getenv('WIP_FOLDER')

import os

def get_video_duration(video_filename):
    """
    Returns the video duration of the provided video file.
    """
    return VideoFileClip(video_filename).duration 

def rescale_video(origin_filename, output_width = 1920, output_height = 1080, output_filename = 'scaled.mp4'):
    """
    This method was created to rescale videos upper to 1920x1080 or 1080x1920. This is,
    when a 4k video appears, we simplify it to 1080p resolution to work with only that
    resolution.

    This method is used in the script-guided video generation. Please, do not touch =).
    """
    # We only want to accept 16/9 or 9/16 by now, so:
    if not (output_width == 1920 and output_height == 1080) and not (output_width == 1080 and output_height == 1920):
        print('Sorry, not valid input parameters.')
        return
    
    SCALE_WIDTH = 16
    SCALE_HEIGHT = 9
    if output_width == 1080 and output_height == 1920:
        SCALE_WIDTH = 9
        SCALE_HEIGHT = 16

    clip = VideoFileClip(origin_filename)

    width = clip.w
    height = clip.h

    # We avoid things like 1927 instead of 1920
    new_width = width - width % SCALE_WIDTH
    new_height = height - height % SCALE_HEIGHT

    proportion = new_width / new_height

    if proportion > (SCALE_WIDTH / SCALE_HEIGHT):
        print('This video has more width than expected. Cropping horizontally.')
        while (new_width / new_height) != (SCALE_WIDTH / SCALE_HEIGHT):
            new_width -= SCALE_WIDTH
    elif proportion < (SCALE_WIDTH / SCALE_HEIGHT):
        print('This video has more height than expected. Cropping vertically.')
        while (new_width / new_height) != (SCALE_WIDTH / SCALE_HEIGHT):
            new_height -= SCALE_HEIGHT

    print('Final: W' + str(new_width) + ' H' + str(new_height))
    clip2 = clip.crop(x_center = floor(width / 2), y_center = floor(height / 2), width = new_width, height = new_height)
    
    # Force output dimensions
    if new_width != output_width:
        print('Forcing W' + str(output_width) + ' H' + str(output_height))
        clip2 = clip2.resize(width = output_width, height = output_height)

    # This fixes the problem of rewriting over an existing video
    clip2.write_videofile(WIP_FOLDER + 'scaled.mp4', codec = 'libx264', audio_codec = 'aac', temp_audiofile = WIP_FOLDER + 'temp-audio.m4a', remove_temp = True)
    os.remove(origin_filename)
    os.rename(WIP_FOLDER + 'scaled.mp4', output_filename)
    
    return True


def extract_audio_from_video(original_filename, output_filename):
    """
    Extracts the audio from the provided 'original_filename' and stores it as the
    'output_filename' audio file.
    """
    VideoFileClip(original_filename).audio.write_audiofile(output_filename)

def detect_scenes(video_filename):
    """
    This method detects the different scenes that are in the provided 'video_filename'
    and returns an array containing 'start' and 'end' elements. Each of those contains
    a 'second' and 'frame' field.

    TODO: This link https://www.scenedetect.com/docs/latest/api.html#example says that
    there is a backend 'VideoStreamMoviePy' class to process it with moviepy. This is
    interesting to use in a clip, and not in a filename.
    """
    # This comes from here: https://www.scenedetect.com/
    # Other project: https://github.com/slhck/scenecut-extractor (only ffmpeg)
    from scenedetect import SceneManager, open_video, ContentDetector

    video = open_video(video_filename, backend = 'moviepy')
    scene_manager = SceneManager()
    scene_manager.add_detector(ContentDetector(threshold = 20))  # 27 is recommended
    scene_manager.detect_scenes(video)

    scenes = []
    for scene in scene_manager.get_scene_list():
        scenes.append({
            'start': {
                'second': scene[0].get_seconds(),
                'frame': scene[0].get_frames()
            },
            'end': {
                'second': scene[1].get_seconds(),
                'frame': scene[1].get_frames()
            }
        })

    return scenes
    
    
    
    
    # This code below is working but it is quite old
    
    from scenedetect import VideoManager, SceneManager, StatsManager
    from scenedetect.detectors import ContentDetector
    from scenedetect.scene_manager import save_images, write_scene_list_html

    video_manager = VideoManager([video_filename])
    stats_manager = StatsManager()
    scene_manager = SceneManager(stats_manager)

    scene_manager.add_detector(ContentDetector(threshold = 30))
    video_manager.set_downscale_factor()

    video_manager.start()
    scene_manager.detect_scenes(frame_source = video_manager)

    scenes = scene_manager.get_scene_list()
    print(f'{len(scenes)} scenes detected!')

    save_images(
        scenes,
        video_manager,
        num_images = 1,
        image_name_template = '$SCENE_NUMBER',
        output_dir = 'scenes')

    for scene in scenes:
        start, end = scene

        # your code
        print(f'{start.get_seconds()} - {end.get_seconds()}')

    return scenes


# This is to work with abruptness and sooness (https://moviepy.readthedocs.io/en/latest/ref/videofx/moviepy.video.fx.all.accel_decel.html)
    
# I previously had 4.4.2 decorator for moviepy. I forced 4.0.2 and it is apparently working


"""
Interesting:
- https://www.youtube.com/watch?v=Ex1kuxe6jRg (el canal entero tiene buena pinta)
- https://www.youtube.com/watch?v=m6chqKlhpPo Echarle un vistazo a ese tutorial
- https://zulko.github.io/moviepy/ref/videofx.html
- https://stackoverflow.com/questions/48491070/how-to-flip-an-mp4-video-horizontally-in-python
"""




from random import randint
from dotenv import load_dotenv

import datetime
import os

load_dotenv()

WIP_FOLDER = os.getenv('WIP_FOLDER')

def create_tmp_filename(filename):
    """
    Returns a temporary file name with 'WIP_FOLDER' prefix and a timestamp suffix.
    """
    delta = (datetime.now() - datetime(1970, 1, 1))
    aux = filename.split('.')
    # TODO: Issue if no extension provided
    return WIP_FOLDER + aux[0] + '_' + str(int(delta.total_seconds())) + str(randint(0, 10000)) + '.' + aux[1]


from yta_general_utils.tmp_processor import create_tmp_filename
from yta_general_utils.video_processor import extract_audio_from_video
from moviepy.editor import VideoFileClip, CompositeVideoClip, CompositeAudioClip, AudioFileClip
from yta_general_utils.experimental.tiktok import download_tiktok_video, process_tiktok_video_url

import numpy as np
import os
import random
import cv2
import math

def build_short_from_tiktok_video(tiktok_url, overlay_video_filename):
    """
    This method receives a tiktok video url and creates a new video to be shared as short
    that contains the original video and an overlay gameplay or content to customize it.
    """
    OVERLAY_CROPPED_VIDEO = create_tmp_filename('overlay_cropped.mp4')
    OVERLAY_RESCALED_VIDEO = create_tmp_filename('overlay_rescaled.mp4')
    
    # Check if exist and has been processed to avoid processing it again
    tiktok_video_data = process_tiktok_video_url(tiktok_url)
    # database_tiktok_video = db_select('tiktok_videos', 'video_id', tiktok_video_data['video_id'])
    # if database_tiktok_video and database_tiktok_video['status'] == 'generated':
    #     print('The video ' + tiktok_video_data['video_id'] + ' from user @' + database_tiktok_video['username'] + ' has already been uploaded. Stopping process...')
    #     return
    
    output_video_filename = tiktok_video_data['username'] + '_' + tiktok_video_data['video_id'] + '_one.mp4'

    # *. Download tiktok video
    downloaded_video_filename = download_tiktok_video(tiktok_url)
    main_video = VideoFileClip(downloaded_video_filename)
    
    if main_video.duration >= 60:
        print('Sorry, main video is 60 seconds or longer...')
        return

    # 1. Rescale overlay to main_video width
    overlay_video = VideoFileClip(overlay_video_filename)
    random_start = random.randint(0, (int) (overlay_video.duration - main_video.duration))
    overlay_video = overlay_video.subclip(random_start, random_start + main_video.duration)
    overlay_video.write_videofile(OVERLAY_CROPPED_VIDEO)

    resize(OVERLAY_CROPPED_VIDEO, OVERLAY_RESCALED_VIDEO, new_width = main_video.w)
    overlay_rescaled_video = VideoFileClip(OVERLAY_RESCALED_VIDEO)

    # 2. Join videos
    final_clip = CompositeVideoClip([main_video.set_position((0, -100)), overlay_rescaled_video.set_position('bottom')])
    final_clip.write_videofile(output_video_filename, fps = 30)

    # 3. Store in database that video as created
    # db_insert('tiktok_videos', ['video_id', 'username', 'status'], [tiktok_video_data['video_id'], tiktok_video_data['username'], 'generated'])

    #  TODO: Remove files
    try:
        os.remove(OVERLAY_CROPPED_VIDEO)
        os.remove(OVERLAY_RESCALED_VIDEO)
        #os.remove(downloaded_video_filename)
    except:
        print('Error deleting files')

def resize(video_filename, output_filename = 'video_resized.mp4', new_width = None, new_height = None):
    if not new_width and not new_height:
        return
    
    clip = VideoFileClip(video_filename)

    if new_width and not new_height:
        new_height = math.ceil((clip.h * new_width) / clip.w)
    elif new_height and not new_width:
        new_width = math.ceil((clip.w * new_height) / clip.h)

    # moviepy resize method does not allow odd numbers
    if new_width % 2 != 0:
        new_width -= 1
    if new_height % 2 != 0:
        new_height -= 1
    
    clip.resize(width = new_width, height = new_height).write_videofile(output_filename)

def deeply_process_green_screen(original_filename, output_filename):
    """
    I receive a green screen transition 'original_filename' and I process it to turn
    that green screen into a black background to, later, improve processing it with
    mask filter.

    This method is useful for clean transitions (maybe with not many colors). This
    need more testing.
    """
    TMP_VIDEO_FILE = create_tmp_filename('tmp_deep_transition_without_audio.avi')
    TMP_AUDIO_FILE = create_tmp_filename('tmp_deep_transition_audio.mp3')

    # TODO: Sorry but this time output must be .avi, thank you
    if not output_filename.endswith('.avi'):
        print('Sorry, output must be .avi')
        exit()

    cap = cv2.VideoCapture(original_filename)
    # grab one frame
    _, frame = cap.read()
    h,w = frame.shape[:2]
    h = int(h)
    w = int(w)
    
    res = (w, h)
    # Videowriter to build the output
    fourcc = cv2.VideoWriter_fourcc(*'XVID')
    out = cv2.VideoWriter(TMP_VIDEO_FILE, fourcc, 30.0, res)
    
    # We only git the most common color in first frame (it should be the green)
    green_screen_color = None
    done = False
    while not done:
        # Get the current frame
        ret, img = cap.read()
        if not ret:
            done = True
            continue

        # Turn into HSV
        hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
        h,s,v = cv2.split(hsv)

        # I choose manually (only in the first frame) the 
        biggest = -1
        if green_screen_color == None:
            # Get unique colors in frame
            unique_colors, counts = np.unique(s, return_counts=True)

            for a in range(len(unique_colors)):
                if counts[a] > biggest:
                    biggest = counts[a]
                    green_screen_color = int(unique_colors[a])

        # This margin was manually chosen (according to results)
        margin = 70
        mask = cv2.inRange(s, green_screen_color - margin, green_screen_color + margin)

        # Smooth out the mask and invert
        kernel = np.ones((3,3), np.uint8)
        mask = cv2.dilate(mask, kernel, iterations = 1)
        mask = cv2.medianBlur(mask, 5)
        mask = cv2.bitwise_not(mask)

        # Crop out the image and fill with color (black) to make it work well
        #crop = np.full_like(img, [0, 255, 0])
        crop = np.zeros_like(img)
        # I think here is filling with BLACK, maybe I can fill it with pure GREEN
        crop[mask == 255] = img[mask == 255]
        out.write(crop)

    # Free resources
    cap.release()
    out.release()

    # Append audio we lost
    extract_audio_from_video(original_filename, TMP_AUDIO_FILE)
    append_audio(TMP_VIDEO_FILE, TMP_AUDIO_FILE, output_filename)

    # Here we have the file stored with the green background now as black background
    # TODO: Maybe we can use that mask to get a transparency layer with another library
    # as opencv does not support it (as I read)
    return True

def append_audio(video_filename, audio_filename, output_filename = 'with_audio.mp4'):
    """
    Receives a video, replaces its audio with the audio in 'input_audio' file
    and creates a new video with the 'output_video' name.
    """
    # TODO: Do this with an array of accepted extensions, please
    if not video_filename.endswith('.mp4') and not video_filename.endswith('.mov') and not video_filename.endswith('.avi') and not video_filename.endswith('webm'):
        video_filename += '.mp4'
    
    # TODO: Create an 'accepted_video_extensions' and 'accepted_audio_extensions' and use them
    if not audio_filename.endswith('.mp3') and not audio_filename.endswith('.wav') and not audio_filename.endswith('.m4a'):
        audio_filename += '.mp3'

    if output_filename == video_filename:
        print('Output video name cannot be the same as input!')
        return None
    
    # TODO: Investigate a way of appending with ffmpeg because it
    # seems to be so much faster than moviepy

    #from subprocess import run
    #parameters = ['ffmpeg', '-i', video_filename, '-i', audio_filename, '-c:v', 'libvpx', '-crf', '15', #'-b:v', '1M', 'copy', output_filename]
    #run(parameters)
    #ffmpeg -i video.mp4 -i audio.m4a -c copy output.mp4


    videoclip = VideoFileClip(video_filename)
    # This is a fix of bad audio ending due to ffmpeg error with moviepy
    # See this: https://github.com/Zulko/moviepy/issues/1936
    #audiounfixedclip = AudioFileClip(input_audio)
    #audiounfixedclip = audiounfixedclip.subclip(0, audiounfixedclip.duration)
    #audioclip = CompositeAudioClip([audiounfixedclip])
    audioclip = CompositeAudioClip([AudioFileClip(audio_filename)])
    videoclip.audio = CompositeAudioClip([audioclip])

    videoclip.to_videofile(
        output_filename,
        #codec = "libx264",
        #temp_audiofile = WIP_FOLDER + 'temp-audio.m4a',
        #remove_temp = True,
        #audio_codec = 'aac' # pcm_s16le or pcm_s32le 'aac' previously
    )


from moviepy.editor import AudioFileClip, VideoFileClip
from pathlib import Path

def write_file(text, filename):
    """
    Writes the provided 'text' in the 'filename' file. It replaces the previous content
    if existing.
    """
    f = open(filename, 'w', encoding = 'utf8')
    f.write(text)
    f.close()

def file_exists(filename):
    """
    Checks if the provided 'filename' exist and is a file.
    """
    return Path(filename).is_file()

def file_is_audio_file(filename):
    try:
        AudioFileClip(filename)
    except:
        return False
    return True

def file_is_video_file(filename):
    try:
        VideoFileClip(filename)
    except:
        return False
    return True